define(["qlik", "jquery", "qvangular", "text!./style.css", "./settings-definition", "text!./template.html","./stretch-dynamic-input"], function (qlik, $,qvangular, cssContent,settingsDefinition, template) {
	'use strict';
	$("<style>").html(cssContent).appendTo("head");
	
	// Add app directives
	addAppDirectives(qvangular);
	
	// run
	return {
		template: template,
		initialProperties: {
			qHyperCubeDef: {
				qDimensions: [],
				qMeasures: [],
				qInitialDataFetch: [{
					qWidth: 50,
					qHeight: 50
				}]
			},
			user: {
				qStringExpression: "=OSUser()"
			},
		},
		definition: settingsDefinition,
		support: {
			snapshot: true,
			export: true,
			exportData: true
		},
		paint: function (){
			//bind the table to the scope for intial load
			if (!this.$scope.table) {
				this.$scope.table = qlik.table(this);
			}

			return qlik.Promise.resolve();
		},
		controller: ['$scope', '$timeout', function ($scope, $timeout) {

			try{
				console.log("Test version");
				



				//get current app
				var app = qlik.currApp(this);				
				configureAppWatches();
				
								//debug console
				if ($scope.layout.props.showDebug) {
					console.log("Scope", $scope);
					console.log("component", $scope.component);
					console.log("App", app, app.model.applyPatches);
					console.log("Qlik", qlik, $scope);
					console.log('layout', $scope.layout);
					console.log("qHyperCube Size", $scope.layout.qHyperCube.qSize);
					console.log("test",$scope.backendApi);
				}	

				
				if(validConfig()){
					//GUI variables
					initAppGUIVariables();

					configureProxySettings();

					initInputDefinition();

					consoleAuthSetting();

					if ($scope.layout.props.showDebug) {
						//var dataPages = $scope.layout.qHyperCube.qDataPages;
						console.log("DataPages", $scope.layout.qHyperCube.qDataPages);
					}

					//Data update
					configureDataUpdateBind();

					//get initial data
					console.log("initial data");
					validateAndAddDataToInputTable();
					
					$scope.appMode = getAppMode();
					
					//TODO rewrite with using qliks build in event listener if they exists. 
					$("button").click(function(){
						setTimeout(function() {
							updateAppMode();
						}, 150); // Delay of 500 milliseconds
					});
				}
				
			
			} catch (ex){
				handleError(ex);				
			}


			
			// Controller functions
			
			function configureAppWatches(){
				//watch configuration

				//watch input definitions
				$scope.$watch('layout.props.inputDefinitionsArray', function (newval, oldval) {
					if ($scope.layout.props.showDebug) {console.log("new definition", newval, oldval, objectsAreEqual(newval, oldval)); }
					if(!objectsAreEqual(newval, oldval)){
						try{
							initInputDefinition();
							validConfig();
							console.log("watch inputDefinitionsArray", newval, oldval, objectsAreEqual(newval, oldval));
							addInputDataToTable();	
						} catch (ex){ handleError(ex);	}
					}

				},true);
				
				//watch service settings
				$scope.$watch('layout.props.service', function (newval, oldval) {
					if ($scope.layout.props.showDebug) {console.log("new settings", newval, oldval,objectsAreEqual(newval, oldval)); }
					if(!objectsAreEqual(newval, oldval)){
						try{
							validConfig();
							configureProxySettings();
							console.log("watch service", newval, oldval,objectsAreEqual(newval, oldval));
							addInputDataToTable();						

						} catch (ex){ handleError(ex);	}
					}
				},true);
			
			}
			
			function objectsAreEqual(object1,object2){
			 return JSON.stringify(object1) === JSON.stringify(object2);
			}
			
			function consoleAuthSetting(){
				if ($scope.layout.props.service.authentication.isEnabled) {
					console.log("Authentication is enabled of type:" + $scope.layout.props.service.authentication.type)
				} else {
					console.log("Authentication is disabled");
				}			
			}
			
			function initAppGUIVariables(){
				//controls disabling input
				$scope.inputReady = false;
				
				//hide the error widget
				$scope.showErrorWidget = false;

				//button active
				$scope.buttonActive = false;

				//status message
				$scope.loadStatus = $scope.layout.props.statusLoadingLabel;

				//User info
				$scope.lastEdit = $scope.layout.props.statusLoadingLabel;
				$scope.lastAction = $scope.layout.props.statusLoadingLabel;
				
			}
			
			function updateAppMode() {
				$scope.appMode = getAppMode();
				if ($scope.layout.props.showDebug) { 
					console.log("Scope app mode", $scope.appMode);
				}
			}
			
			function getAppMode() {
				//if app mode is edit then edit else analysis
				if ($scope.layout.props.showDebug) { 
					console.log("setting app mode");				
				}

				return qlik.navigation.getMode();
			}
			
			
			function validateAndAddDataToInputTable(){
				if (dataPagesHasValidData($scope.layout.qHyperCube.qDataPages)) {
					//console.log("got data");
					addInputDataToTable();
				} else {
					console.log("No data return from Qlik");
					$scope.buttonActive = false;
					setMessageToNoData();
				}
			}
			
			function configureDataUpdateBind(){
				$scope.component.model.Validated.bind
				(function () {				
					if ($scope.layout.props.showDebug) { console.info('Validated, new data'); }
														
					
					validateAndAddDataToInputTable();

				});	
				
				// invalid is called before reload of data
				$scope.component.model.Invalidated.bind(function(){					
					
					if ($scope.layout.props.showDebug) {console.log("start data invalid");};
					if(!!$scope.activeInputInfo && !!$scope.inputContainer){
						let activeValue = getInputFromActiveInputInfo($scope.inputContainer, $scope.activeInputInfo);
						if(!!activeValue){
							$scope.activeInputInfo.input = activeValue;
							if ($scope.layout.props.showDebug) {console.log("Stored active input", $scope.activeInputInfo);};
						} else {
							if ($scope.layout.props.showDebug) {console.log("Found no stored data", activeValue);};
						}
					}
					if ($scope.layout.props.showDebug) {console.log("invalid data", "activeInfo", $scope.activeInputInfo, $scope.inputContainer,!!$scope.activeInputInfo && !!$scope.inputContainer );};
				});
			}
			
			function configureProxySettings(){
				//get proxy settings
				var proxyBaseUrl = $scope.layout.props.service.proxyBaseUrl; 
				var proxyPort = $scope.layout.props.service.proxyPort;
				var proxyEndpoint = $scope.layout.props.service.proxyEndpoint + '/' + $scope.layout.props.service.extensionID;

				//build connection string
				if (proxyPort == "80" || proxyPort == "443") {
					var proxyUrl = proxyBaseUrl + proxyEndpoint;
				} else {
					var proxyUrl = proxyBaseUrl + ':' + proxyPort + proxyEndpoint;
				}
				
				$scope.proxyUpdateComment = proxyUrl +'/updaterecord';
				$scope.proxyGetNewComments = proxyUrl + '/getnewestrecords';

				$scope.proxyInteractiveAuth = proxyUrl + '/about';
				
				$scope.proxyBuildViews = proxyUrl + '/buildallviews';
			}
			
			function initInputDefinition(){					
				//store input template for use
				//$scope.inputDefinition = $scope.layout.props.inputDefinitionsArray.map(({ id, header, placeholder, type, values, attributes, optionalAttributes }) => ({ id, header, placeholder, type, values, attributes, optionalAttributes }));
				$scope.inputDefinition = $scope.layout.props.inputDefinitionsArray.map(function(element){ 
					return { id: element.id, header: element.header, placeholder: element.placeholder, columnIsDisabledExpression: element.columnIsDisabledExpression , type: element.type, values: GetValuesFromFixedOrDynamic(element), layout: element.layout, attributes: element.attributes, optionalAttributes: element.optionalAttributes, valuesType: element.valuesType, dynamicValueExpression: element.dynamicValueExpression  }
				});
				
				
				if ($scope.layout.props.showDebug) { console.log("Found mapped input definition", $scope.inputDefinition ) }

				//setup the initial input definition and container
				$scope.inputContainer = angular.copy($scope.inputDefinition);	

			}
			
			//Todo add check for missing keys and values, and display missing keys or values at front end
			function GetValuesFromFixedOrDynamic(element){
				

				if(!!element && element.valuesType == 'fixed' ){
					return element.values;				
				}
				
				if(!!element && element.valuesType == 'dynamic' && !!element.dynamicValueExpression){

					const dynamicValues =	element.dynamicValueExpression.split('|').map(item => {
						const [key, value] = item.split('~');
						return { key, value };
				  	});

					//if ($scope.layout.props.showDebug) { console.log("Returning dynamic value", dynamicValues) }
					return dynamicValues;
				}
				
				//default fallback
				return element.values;			
			}


			//validate config
			function validConfig(){
			
				var props = $scope.layout.props;
				
				//services
				
				var serviceSettingsText = ", add it under service settings";
				
				/*if(isIE()){
					throw makeError("Internet Explorer is not supported","Please Update to Edge or Chrome, browsers supported by the current Qlik Sense release is also supported","Configuration");
				}*/
				
				if(isStringDefined(props.inputDefinition)){
					throw makeError("Old configuration of input definition found","Please Update in the Input Definition and delete in the table and data settings, then reload the page","Configuration");
				}
				
				if(!props.service){
					throw makeError("Please update the service settings","No Writebase service settings is defined"+serviceSettingsText,"Configuration");
				}
				
				if(!isStringDefined(props.service.proxyBaseUrl)){
					throw makeError("Please add the url to the service","No Writebase service url is defined"+serviceSettingsText,"Configuration");
				}
				
				if(!isStringDefined(props.service.extensionID)){
					throw makeError("Please a solution Id","No solution extension id is defined"+serviceSettingsText,"Configuration");
				}
				
				if(!isStringDefined(props.service.proxyEndpoint)){
					throw makeError("Please add a proxy end point","No proxy endpoint is defined"+serviceSettingsText,"Configuration");
				}
				
				if(!props.service.authentication){
					throw makeError("Please update autthentication ","No authentication info defined"+serviceSettingsText,"Configuration");
				}
				
				if(!isStringDefined(props.service.authentication.type)){
					throw makeError("Please select autthentication type","No authentication type is defined"+serviceSettingsText,"Configuration");
				}
				
				if(props.service.authentication.isEnabled == undefined){
					throw makeError("Please select if autthentication is enabled","No authentication enabled/disabled is defined"+serviceSettingsText,"Configuration");
				}
				
				if(props.service.authentication && isStringDefined(props.service.authentication.type) && props.service.authentication.type == "HEADER" && !isStringDefined(props.service.authentication.apiKey)){
					throw makeError("Please add a API key","Authentication is set to Header with no API key defined"+serviceSettingsText,"Configuration");
				}
				
				//table settings
				
				if(!isStringDefined(props.dataColumnName)){
					throw makeError("Please add the data field name to the table and data setting","No data field name is defined"+serviceSettingsText,"Configuration");
				}
				
				
				if(!isStringDefined(props.keySeperator)){
					throw makeError("Please add the key seperator to the table and data setting","No data key seperator is defined"+serviceSettingsText,"Configuration");
				}
				
				
				if(getKeyColumnsInformationFromDimensionsInfo($scope.layout.qHyperCube.qDimensionInfo).length == 0){
					throw makeError("Please ensure that atleast one dimension have partOfKey checked","No dimesions is set as key, add it ot the relevant dimension(s). Reload the page if required","Configuration");
				}
				
				return true;			
			}
			
			function isStringDefined(configValue){
				return !!configValue && configValue.length > 0;
			}
			
			
			function isIE() {
				const ua = window.navigator.userAgent; //Check the userAgent property of the window.navigator object
				const msie = ua.indexOf('MSIE '); // IE 10 or older
				const trident = ua.indexOf('Trident/'); //IE 11

				return (msie > 0 || trident > 0);
			}

			//update sort order
			function setSortOrder ( scope, columnIndex ) {
				//column first
				var sortOrder = [columnIndex];
				//append the rest
				scope.layout.qHyperCube.qEffectiveInterColumnSortOrder.forEach( function ( currentColumnIndex ) {
					if ( currentColumnIndex !== sortOrder[0] ) {
						sortOrder.push( currentColumnIndex );
					}
				} );
				
				var sortOrder
				
				scope.backendApi.applyPatches( [{
					'qPath': '/qHyperCubeDef/qInterColumnSortOrder',
					'qOp': 'replace',
					'qValue': '[' + JSON.stringify(sortOrder) + ']'
				}], true );
			}

			//reverste sort order
			function reverseOrder ( scope, columnIndex ) {
				var hyperCube = scope.layout.qHyperCube;
				var numberOfDimension = hypercube.qDimensionInfo.length;
				
				if(columnIndex < numberOfDimension){
					var columnReverseSortState = hyperCube.qDimensionInfo[columnIndex].qReverseSort;
					var apiPath = 'qDimensions/' + columnIndex;
				} else {
					var measureIndex = columnIndex - numberOfDimension
					var columnReverseSortState = hyperCube.qMeasureInfo[measureIndex].qReverseSort;
					var apiPath = 'qMeasures/' + (measureIndex);
				}

				scope.backendApi.applyPatches( [{
					'qPath': '/qHyperCubeDef/' +
					apiPath +
					'/qDef/qReverseSort',
					'qOp': 'replace',
					'qValue': ( !columnReverseSortState ).toString() //set to reverse state
				}], true );
			}
			
			

			//gui Function
			
			$scope.setSortOrder = function(headElement){
				console.log("setting order for", headElement);

				// Scroll to top of extension root div
				var scrollTopOfElement = angular.element("#qloud_cover_writeback")[0];
				scrollTopOfElement.scrollTop = 0;

				if($scope.layout.qHyperCube.qEffectiveInterColumnSortOrder[0]==headElement.col){
					headElement.reverseOrder();
				} else {
					headElement.orderBy();
				}				
				
			}

			//save input when inputfield is deselected for multiple input type
			$scope.onInputBlur = function (row, colmumnKey, input, rowData, inputDefinition) {
				let rowKey = getRowKeyFromRow(row);
				
				
				//clear active info
				$scope.activeInputInfo = null;
				
				if ($scope.layout.props.showDebug) {
					console.log("Debluring", rowKey, colmumnKey, input, rowData, $scope.activeInputInfo);
					console.log("New input", "rowKey", rowKey, "columnKey", colmumnKey, "input", input, "rowData", rowData);
				}

				//TODO reverse logic
				if ((input != undefined) && !!rowKey && !!colmumnKey) {
					
				
					//skip update if matches the current or update if input type is range
					if ( $scope.oldInput == input && (!!inputDefinition && inputDefinition.type != "range")	) {	
						console.log("no new input, skipping update");
						return;
					}	

						
					//update data if all input is valid
					updateData(rowKey, colmumnKey, input, rowData);
					return;			
						
					
				} else {
					console.log("got empty update ignoring");
				}

			};
			
			//save data if for single input type
			$scope.onInputChange = function(row, colmumnKey, input, rowData, oldValue) {
				let rowKey = getRowKeyFromRow(row);
				if ($scope.layout.props.showDebug) {
					console.log("onchange", rowKey, colmumnKey, input, rowData, oldValue);
					console.log("New input", "rowKey", rowKey, "columnKey", colmumnKey, "input", input, "rowData", rowData);
				}
				
				//console.log("onchange", rowKey, colmumnKey, input, rowData, oldValue );
				
				if($scope.inputReady == false){
					if ($scope.layout.props.showDebug) { 
					console.log("Input is initlizing, ignoring");
					}
					return;
				}
	
				if ((input != undefined) && !!rowKey && !!colmumnKey && (oldValue == undefined || (!!oldValue && oldValue != ""))  ) {					
					//update data if all input is valid
					updateData(rowKey, colmumnKey, input, rowData);					
					
				} else {
					if ($scope.layout.props.showDebug) { 
					console.log("got empty update ignoring");
					}
					return;
				}
			
			}

			//capture old input to track changes
			$scope.onInputFocus = function (row, colmumnKey, input, rowIndex) {
				let rowKey = getRowKeyFromRow(row);
				
				
				//set active info
				
				$scope.activeInputInfo = {'rowKey': rowKey, 'columnKey': colmumnKey, 'input': input, 'rowIndex': rowIndex};
				
				
				if ($scope.layout.props.showDebug) { console.log("Has focus", "row", row, "columnKey", colmumnKey, "input", input,"Rowkey", rowKey, 'activeInputInfo', $scope.activeInputInfo) }
				
				
				
				$scope.oldInput = input;
			};

			//autoexpand input to wrap input
			$scope.autoExpand = function (element) {
				function resize() {
					if(element.scrollHeight>element.clientHeight){
						element.style.height = element.scrollHeight + 'px';
					}
				}
				window.setTimeout(resize, 0);
				//$timeout(resize, 0, false);					
				resize();
			}

			//expand all
			$scope.autoExpandAllofType = function (parentId, elementType) {
				//get all text areas
				var elements = Array.from(document.getElementById(parentId).getElementsByTagName(elementType));
	
				function resizeAll() {
					elements.map(function (element) {
						
						//element.style.height = 'auto';
						//check if height needs adjustment
						if(element.scrollHeight>element.clientHeight){
							element.style.height = element.scrollHeight + 'px';
						}
					});
				}
				
				window.setTimeout(resizeAll, 0);
				
			}
			 

			//action function
			$scope.actionButton = function () {
				$scope.loadStatus = $scope.layout.props.actionStatusLabel;

				var keys = getKeysFromDataPages($scope.layout.qHyperCube.qDataPages, getKeyColumnIndex($scope.layout.qHyperCube.qDimensionInfo));
				var baseKey = getBaseKey(keys);

				var keySeperator = $scope.layout.props.keySeperator;
				var actionKey = $scope.layout.props.actionKey;

				if ($scope.layout.props.showDebug) { console.log("performing action for baseKey", baseKey); }

				var recordInfo = { Type: "action", action: "actionButton" };
				var guiInfo = { source: "QlikSense" }

				var jsonMessage = {
					key: baseKey + keySeperator + actionKey,
					input: 'actionButton',
					recordInfo: JSON.stringify(recordInfo),
					guiInfo: JSON.stringify(guiInfo)
				};

				postMessageToProxy($scope.proxyUpdateComment, jsonMessage).then(function (message) {
					console.log("got message from proxy", message.response);
					$scope.loadStatus = $scope.layout.props.statusReadyLabel;
					$scope.lastAction = new Date(); //set update to now
					$scope.$apply();

				});
			}
			$scope.setCalculatedStyle = function (columnIndex, rowIndex) {
				var styleObject = {};

				//get expression, 0 is the index on the font color
				var color = getCellStyleColor(columnIndex, rowIndex, 0);
				var backgroundColor = getCellStyleColor(columnIndex, rowIndex, 1);
				var columnFontWeight = getColumnStyle(columnIndex, "styleFontWeight");

				//return "{'color':'" & color & "'}";

				//check if color is set
				if (!!color) {
					styleObject.color = color;
				}
				if(!!backgroundColor){
					styleObject.backgroundColor = backgroundColor;
				}
				if (!!columnFontWeight) {
					styleObject["font-weight"] = columnFontWeight;
				}
				//console.log("style", styleObject, columnIndex, rowIndex);

				return styleObject;

			}
			
			$scope.calcTableIndex = function (columnIndex, rowIndex) {
				//limits the datat to 1M row for tab order
				return rowIndex + 1 + columnIndex * 1000000;
			}
			
			$scope.rowIsDisabled = function(columnIndex, rowIndex){
				
				//get expression, 2 is the index on the active dimension
				var isDisabled = getCalculatedExpression(0, rowIndex,2);

				
				if(!!isDisabled){
					//console.log(rowIndex, isDisabled, JSON.parse(isDisabled.toLowerCase()) === true || JSON.parse(isDisabled.toLowerCase()) === -1 );
					//return JSON.parse(isDisabled.toLowerCase()) === true || JSON.parse(isDisabled.toLowerCase()) === -1;
					return castSetExpressionToBoolean(isDisabled);
				
				}
				
				return false;
			}
			
			$scope.columnIsDisabled = function(columnIsDisabledExpression){
				
				//get expression, 2 is the index on the active dimension
				var isDisabled = columnIsDisabledExpression;
				if(!!isDisabled){
					//console.log(isDisabled, JSON.parse(isDisabled.toLowerCase()) === true || JSON.parse(isDisabled.toLowerCase()) === -1 );
					//console.log("Calling columns is disabled function with expression ", columnIsDisabledExpression, isDisabled, JSON.parse(isDisabled.toLowerCase()) === true || JSON.parse(isDisabled.toLowerCase()) === -1)
					//return JSON.parse(isDisabled.toLowerCase()) === true || JSON.parse(isDisabled.toLowerCase()) === -1;
					return castSetExpressionToBoolean(isDisabled)
				}
				
				return false;
			}
			
			$scope.cellIsDisabled = function(columnIsDisabledExpression, columnIndex, rowIndex ) {
				var isDisabled = $scope.rowIsDisabled(columnIndex, rowIndex) || $scope.columnIsDisabled(columnIsDisabledExpression);
				//console.log("Function is disabled: ", isDisabled, columnIsDisabledExpression, columnIndex, rowIndex);
				return isDisabled;
			}
		
			function castSetExpressionToBoolean(setExpression) {

				var expression = setExpression.trim().toLowerCase();
				return JSON.parse(expression) === true || JSON.parse(expression) === -1;

			}
			
			
			// get more table data
			$scope.tableGetMoreData = function () {
				var originalKeys = getKeysFromDataPages($scope.layout.qHyperCube.qDataPages, getKeyColumnIndex($scope.layout.qHyperCube.qDimensionInfo));
				
				// Scroll to top of extension root div
				var loadButtonElement = angular.element("button.lui-button.more")[0];
				//handle that the button might not be shown if the list is below the initial load limit
				if(!!loadButtonElement){
					loadButtonElement.classList.add('load-more-loading');
				}

				// create table listener which we unbind later
				var tableGetMoreDataListener = function () {
					if ($scope.layout.props.showDebug) { console.log('table ondata event') };
					var updatedKeys = getKeysFromDataPages($scope.layout.qHyperCube.qDataPages, getKeyColumnIndex($scope.layout.qHyperCube.qDimensionInfo));

					// might filter duplicate rows which we want to keep. TODO: Maybe count original keys, and remove upto that amount in new keys?
					var newKeys = updatedKeys.filter(function (el) {
						return originalKeys.indexOf(el) < 0;
					});

					addInputDataToTable(updatedKeys);
					$scope.table.OnData.unbind(tableGetMoreDataListener);  //unregister the listener when no longer notification is needed.
					loadButtonElement.classList.remove('load-more-loading');
				};

				// bind listener OnData
				$scope.table.OnData.bind(tableGetMoreDataListener); //bind the listener


				// Call get more data, listener should then trigger and call service endpoint
				//console.log("loading more data");
				unBlurActiveInput();
				//console.log("table",$scope.table, $scope.table.getMoreData);
				$scope.table.getMoreData();
			}

			//style cell
			function getCellStyleColor(columnIndex, rowIndex, expressionIndex) {
				
				var styleData = getCalculatedExpression(columnIndex, rowIndex,expressionIndex);
				
				//check if there is an expression
				if(!!styleData){
					var expectedStart = "RGB(";
					if (styleData.substring(0, expectedStart.length) == expectedStart) {
						//return { "color": styleData.qText };
						return styleData;
					} else {
						console.error("invalid color", styleData);
					}
				}

				//default return - no styles applied
				return null;
			}

			
			function getDataPageSize(dataPages){
				//TODO fix support for larger initial data.
				return dataPages[0].qMatrix.length;
			}
			
			function getDataPageIndex(dataPageSize, rowIndex){
				return (rowIndex - rowIndex % dataPageSize) / dataPageSize
			
			}
			
			function getDataPageRowIndex(dataPageSize,rowIndex){
				return rowIndex % dataPageSize;
			}
			
			function getCalculatedExpression(columnIndex, rowIndex, expressionIndex){
				//check if table is present and the hupercube is loaded
				if(!!$scope.table && !!$scope.table.qHyperCube && !!$scope.table.qHyperCube.qDataPages[0] && !!$scope.table.qHyperCube.qDataPages[0].qMatrix && !!$scope.table.qHyperCube.qDataPages[0].qMatrix.length > 0){
					var dataPageSize = getDataPageSize($scope.table.qHyperCube.qDataPages);
					var dataPageIndex = getDataPageIndex(dataPageSize,rowIndex);
					var dataPageRowIndex = getDataPageRowIndex(dataPageSize, rowIndex);

					if (!!$scope.table.qHyperCube.qDataPages[dataPageIndex] && !!$scope.table.qHyperCube.qDataPages[dataPageIndex].qMatrix && !!$scope.table.qHyperCube.qDataPages[dataPageIndex].qMatrix[dataPageRowIndex] && !!$scope.table.qHyperCube.qDataPages[dataPageIndex].qMatrix[dataPageRowIndex][columnIndex]) {
						var cellData = $scope.table.qHyperCube.qDataPages[dataPageIndex].qMatrix[dataPageRowIndex][columnIndex];
						if (!!cellData && !!cellData.qAttrExps && !!cellData.qAttrExps.qValues[expressionIndex] && !!cellData.qAttrExps.qValues[expressionIndex].qText) {
							var expressionData = cellData.qAttrExps.qValues[expressionIndex];
							return expressionData.qText;

						} else {
							//console.log("no expression data");
						}
					} else {
						//console.error("no cell data", columnIndex, rowIndex);
					}
				}
				return null;
			}


			function getColumnStyle(columnIndex, property) {
				if (!!$scope.table && !!$scope.table.headers && !!$scope.table.headers[columnIndex]) {
					//handle dimension
					if (!!$scope.table.headers[columnIndex].qDimensionInfo) {
						if (!!$scope.table.headers[columnIndex].qDimensionInfo[property]) {
							return $scope.table.headers[columnIndex].qDimensionInfo[property];
						}
						//handle measure
					} else if (!!$scope.table.headers[columnIndex].qMeasureInfo) {
						if (!!$scope.table.headers[columnIndex].qMeasureInfo[property]) {
							return $scope.table.headers[columnIndex].qMeasureInfo[property];
						}
					}
				} else {
					//console.error("no header data", columnIndex);					
				}
				//default return - no styles applied
				return null;

			}


			function unBlurActiveInput(){
				let elem = document.activeElement//document.querySelector(':focus');
				//console.log("active element", elem);
				if(elem ) {
				  elem .blur();
				}
			}
			
			


			//gui helper funtions
			function setMessagesToLoading() {
				$timeout(function () {
					$scope.showErrorWidget = false;
					$scope.lastEdit = $scope.layout.props.statusLoadingLabel;
					$scope.lastAction = $scope.layout.props.statusLoadingLabel;
					$scope.loadStatus = $scope.layout.props.statusLoadingLabel;
				}, 0, false);
			}

			function setMessagesToAuthenticating() {
				//TODO should make a message handler
				handleError(makeError($scope.layout.props.authenticatingLabel,"Authentication Required, pending interactive authentication"));
				/*$timeout(function () {
					$scope.showErrorWidget = false;
					$scope.loadStatus = $scope.layout.props.authenticatingLabel;
					$scope.lastEdit = "";
					$scope.lastAction = "";
				}, 0, false);*/
			}

			function setMessageToNoData() {
				$timeout(function () {
					$scope.showErrorWidget = false;
					$scope.loadStatus = $scope.layout.props.noDataLabel;
					$scope.lastEdit = $scope.layout.props.lastNoChangeLabel;
					$scope.lastAction = $scope.layout.props.lastNoActionLabel;
				}, 0, false);
			}
			
			function handleError(error){
				$timeout(function () {				
					
					//clear gui messages
					$scope.loadStatus = "";
					$scope.lastEdit = "";
					$scope.lastAction = "";
					
					$scope.showErrorWidget = true;
					$scope.errorMessage = formatError(error);
					console.error("Extension failed with error",error);
					
					//throw exception if defined and debug
					if ($scope.layout.props.showDebug) {
						if(!!error.exception){
							//throw error.exception;
						}
					};
					
				}, 0, false);
			}
			
			function clearErrorWidget(){
				$timeout(function () {
					$scope.showErrorWidget = false;
					$scope.errorMessage = null;
				}, 0, false);
			}
			
			
			function makeOrParseError(error, message){
				return formatError(error, message);
			}
			
			function formatError(error, message){
				if(error.hasOwnProperty('friendlyMessage') && error.hasOwnProperty('supportMessage'))	{
					return error;
				} else {
					if(typeof message == "undefined"){
						message = 'Sorry an error occurred';
					}
					return makeError(message,error);
				}
			
			}
			
			function makeError(friendlyMessageText, exceptionText, errorId, supportText){
				
				if(typeof friendlyMessageText == "undefined") {
					friendlyMessageText = "Default Error";
				}

				if(typeof exceptionText !== "undefined") {
					exceptionText = "Technical : " + exceptionText;
				}
				
				if(typeof supportText == "undefined") {
					if(!!$scope.layout.props.statusErrorLabel){
						supportText = $scope.layout.props.statusErrorLabel;
					}
				}				
				
				if(typeof supportText == "undefined") {
					errorId = "Unknown";
				}
				
				return 	{
					type: 'error',
					id: errorId,
					exception: exceptionText,
					friendlyMessage: friendlyMessageText,
					supportMessage: supportText
				}
			}


			//interactive auth

			$scope.openAuthWindow = function () {
				return new Promise(function (resolve, reject) {
					console.log("opening windows");
					var win = window.open($scope.proxyInteractiveAuth);

					if (win == null) {
						console.log("popup disabled / not allowed");
						
						reject(makeError("popup disabled / not allowed","popup disabled / not allowed","popupblocked"));

					} else {

						//time out
						var windowTimeout = setTimeout(function () {
							win.close();
							resolve('closed');
							clearInterval(windowLoadedCheck);
						}, 120000);

						//check for auth complete
						var windowLoadedCheck = setInterval(function () {
							console.log("check for succes at url:", $scope.proxyInteractiveAuth);
							if (win.location == undefined || win.location.href == "about:blank") {
								console.log("window location is undefined or about:blank");
							} else if (!!win.location && !!win.location.href && win.location.href.indexOf($scope.proxyInteractiveAuth) == 0) { //checks if the start url is identically to handle added queries from the proxy like #
								console.log("auth successfull");
								clearTimeout(windowTimeout);
								clearInterval(windowLoadedCheck);
								win.close();
								resolve('success');
							} else {
								console.log("unexpected state", win);
								clearTimeout(windowTimeout);
								clearInterval(windowLoadedCheck);
								
								reject(makeError("Authentication failed or timed out without success","Extention encountered an error, was not redirected to the url","authenticationfailed"));
							}

						}, 500);
					}


					/*reject({
						status: xhr.status,
						statusText: xhr.statusText
					});*/
				});
			}


			//data functions
			
			function getInputFromActiveInputInfo(inputData,activeInputInfo){
				let inputColumn = inputData.filter(function(inputColumn){
					return activeInputInfo.columnKey == inputColumn.id
				});
				if ($scope.layout.props.showDebug) {console.log("looking data input data",inputColumn);};
				
				if(inputColumn.length >0){
					if ($scope.layout.props.showDebug) {console.log("found data",inputColumn[0].data[activeInputInfo.rowIndex], inputColumn[0].data, activeInputInfo);};
					return inputColumn[0].data[activeInputInfo.rowIndex];
				}
				return null;
			
			}
			
			function updateInputContainerDataFromStoredInput(inputData,activeInputInfo){

				if(!!activeInputInfo && !!inputData && !!activeInputInfo.input){
					
					let keys = getKeysFromDataPages($scope.layout.qHyperCube.qDataPages, getKeyColumnIndex($scope.layout.qHyperCube.qDimensionInfo));
					let keySepetrator = $scope.layout.props.keySeperator;
					let inputKey = activeInputInfo.rowKey
					
					if(keys.includes(inputKey)){
					
						let inputColumn = inputData.filter(function(inputColumn){
						return activeInputInfo.columnKey == inputColumn.id
						});
						if ($scope.layout.props.showDebug) {console.log("looking for input data to update",inputColumn);};

						if(inputColumn.length >0 && !!activeInputInfo.input){
							if ($scope.layout.props.showDebug) {console.log("found data",inputColumn[0].data[activeInputInfo.rowIndex], inputColumn[0].data, activeInputInfo);};	
							inputColumn[0].data[activeInputInfo.rowIndex] = activeInputInfo.input;
						}
					
					} else {
						if ($scope.layout.props.showDebug) {console.log("key not in keys", keys, inputKey)};
					
					}
					
					
				}			
			}
			
			function getInputFromServerData(serverData,activeInputInfo){
				var keySepetrator = $scope.layout.props.keySeperator;
				let inputKey = activeInputInfo.rowKey + keySepetrator + activeInputInfo.columnKey;
				if ($scope.layout.props.showDebug) {console.log("looking for input in data",inputKey,serverData,activeInputInfo);};
				return serverData.filter(function (inputRecord) { return inputRecord.key ==  inputKey})[0];
			}
			
			function updateInputDataFromStoredInput(serverData,activeInputInfo){
				if(!!serverData && !!activeInputInfo){
					if ($scope.layout.props.showDebug) {console.log("updating server data with stored active input",serverData, activeInputInfo);};
					
					//need to check the table for data
					var inputToUpdate = getInputFromServerData(serverData,activeInputInfo);
					
					if(!!inputToUpdate && !!inputToUpdate.key && !!inputToUpdate.data){
						if ($scope.layout.props.showDebug) {console.log("Retrieving old active input", $scope.activeInputInfo);};
						inputToUpdate.data = activeInputInfo.input;
					} else {
						if ($scope.layout.props.showDebug) {console.log("Found no data input to update", inputToUpdate);};
					
					}
				
				}
				
			}

			//Load data into input
			function addInputDataToTable (keys) {
				try {
					if ($scope.layout.props.showDebug) { console.log("checking table for consistant key"); }
					$scope.inputReady = false;
					$scope.buttonActive = false;
					setMessagesToLoading();

					//$scope.loadStatus = $scope.layout.props.statusLoadingLabel;
					var url = $scope.proxyGetNewComments;
					var keys = keys || getKeysFromDataPages($scope.layout.qHyperCube.qDataPages, getKeyColumnIndex($scope.layout.qHyperCube.qDimensionInfo));
					if ($scope.layout.props.showDebug) { console.log("Keys", keys); }
					
					if( $scope.layout.props.useUniqueKeys ) {
						// Use uniqueKeys
						// ...
					} else {
						// Use baseKey
						var baseKey = getBaseKey(keys);
						url = url + '/' + baseKey;

						if(!baseKey || baseKey.length == 0){
							throw makeError("Please add dimension a valid key dimension","Use a dimension containing a multi-part key seperated by the key seperator (default is '|')","Configuration")
						}

						if ($scope.layout.props.showDebug) {
							console.log("baseKey", baseKey);
							console.log("updating inputcontainer");
						}
					}

					//get newest comments
					getDataFromProxy( url, keys ).then(function (data) {
						if (!!data && data.length > 0 && Array.isArray(data)) {
							if ($scope.layout.props.showDebug) { console.log("Got data", data); }
							if(!!$scope.layout.props.dataColumnName){
								var dataCloumnData = $scope.layout.props.dataColumnName;
							} else {
								var dataCloumnData = "data";
							}
							
							//udpate input data if stored
							//updateInputDataFromStoredInput(data,$scope.activeInputInfo);

							//get definitions
							var inputDefinitions = angular.copy($scope.inputDefinition);
							var keySepetrator = $scope.layout.props.keySeperator;

							//create empty container
							var inputContainer = [];

							inputDefinitions.forEach(function (inputDefinition) {
								//if ($scope.layout.props.showDebug) { console.log("processing inputDefinition", inputDefinition); }
								var keyInputData = [];

								//loop keys
								keys.forEach(function (key) {

									//if ($scope.layout.props.showDebug) { console.log("processing", "key", key, "inputDefinition", inputDefinition); }
									var inputData = data.filter(function (inputRecord) { return inputRecord.key == key + keySepetrator + inputDefinition.id });

									if (!!inputData && inputData.length == 1) {
										var inputRecord = inputData[0];
										//if ($scope.layout.props.showDebug) { console.log("found input data", inputRecord)};
										if(inputRecord[dataCloumnData] == undefined){throw makeError("Please check the data field name","No data in data field '" + dataCloumnData + "'")}
										keyInputData.push(parseDataTypesFromInputType(inputRecord[dataCloumnData],inputDefinition.type) );
									} else {
										if (!!inputData && inputData.length > 1) {
											if ($scope.layout.props.showDebug) { console.log("found multiple records, expects one "); }
										} else {
											//if ($scope.layout.props.showDebug) { console.log("no records, expects one "); }
										}
										
										//get the empty value based on input definition
										
										var emptyValue = getDefaultInputValueOnEmpty(inputDefinition);
										keyInputData.push(emptyValue);
									}
								});

								//update inputContainer
								inputContainer.push(
									{
										header: inputDefinition.header,
										id: inputDefinition.id,
										data: keyInputData

									}
								);
							});

							if ($scope.layout.props.showDebug) { console.log("updating input container", inputContainer); }
							//update gui with data
							$timeout(function () {
								
								// check if uniqueKeys setting is enable AND $scope.inputContainer contains data field, meaning data already exists
								// in which case data is appended, else data is set and replaced
								/*
								if($scope.layout.props.useUniqueKeys && (typeof $scope.inputContainer[0].data !== 'undefined')) {
									$scope.inputContainer.push.apply(inputContainer);
									console.log('inputContainer', inputContainer)
									console.log('$scope.inputContainer', $scope.inputContainer)
								}
								else {
									$scope.inputContainer = inputContainer;
								}
								*/
								updateInputContainerDataFromStoredInput(inputContainer, $scope.activeInputInfo);								
								$scope.inputContainer = inputContainer;
								


								$scope.lastEdit = findLastestTimestampIndata(data) || $scope.layout.props.lastNoChangeLabel;
								$scope.lastAction = getLastActionTimestamp(baseKey, data) || $scope.layout.props.lastNoActionLabel;

								//trigger textarea update after digest
								$timeout(function () {
									$scope.autoExpandAllofType("stretch-Input-table", "textarea");
								}, 2, false);
							}, 0, false);

						} else {
							$timeout(function () {
								console.log("no data or empty array was returned", data);
								$scope.inputContainer = angular.copy($scope.inputDefinition);
								if ($scope.layout.props.showDebug) { console.log("inputContainer", $scope.inputContainer, $scope.inputDefinition); }
								$scope.lastEdit = $scope.layout.props.lastNoChangeLabel;
								$scope.lastAction = $scope.layout.props.lastNoActionLabel;

								//trigger textarea update after digest
								$timeout(function () {
									$scope.autoExpandAllofType("stretch-Input-table", "textarea");
								}, 2, false);
							}, 0, false);

						}
						$timeout(function () {
							//console.log("done");
							$scope.inputReady = true;
							$scope.buttonActive = true;
							$scope.loadStatus = $scope.layout.props.statusReadyLabel;
							$scope.showErrorWidget = false;
							$scope.$apply();
						}, 3, false);

					})
						//handle errors
						.catch(function (error) {
							if ($scope.layout.props.showDebug) { console.error("error in getting data", error); }
							//catch popup blocker
							if (!!error.id && error.id == "popupblocked") {
								setMessagesToAuthenticating();
							} else {
								handleError(makeOrParseError(error,"Error in getting data"));
							}
						}

						);


					//$scope.$apply();

				} catch (ex) {
					handleError(makeOrParseError(ex,"Error in updating table"));
					
				}

			};
			
			function parseDataTypesFromInputType(data, inputType){
				if(inputType == "date"){
					var date = new Date(data.replace(/T\d{2}/,'T00'));
					//console.log("date", data ,'|',data.replace(/T\d{2}/,'T00'),'|', date )
					return date;
				} else if(inputType == "range"){
					return parseInt(data);
				} else if(inputType == "number"){
					 if (Number.isNaN(Number.parseFloat(data))) {
						return 0;
					  }
					  return parseFloat(data)
				} else {
					return data;
				}			
			}
			
			function getDefaultInputValueOnEmpty(inputDefinition){
				
				if(!!inputDefinition && inputDefinition.type == "range"){
					
					if(!!inputDefinition.attributes && inputDefinition.attributes.defaultvalue != undefined){
						return inputDefinition.attributes.defaultvalue;					
					}
								
				}				
				return undefined;
			}

			
			$scope.updateViewsButton = function(){
				$scope.loadStatus = "Updating views";
				if ($scope.layout.props.showDebug) { console.log("Clicked Update Views Button"); }
				updateViews();

			}
			// autocast views function
			function updateViews(){
			
				var inputDefinitions = $scope.layout.props.inputDefinitionsArray;
				if ($scope.layout.props.showDebug) {
				console.log("InputDefinitions", inputDefinitions);
				}
				
				var jsonMessage = inputDefinitions.map(mapJsonToInputDefinitionConfiguration);
				

				 if ($scope.layout.props.showDebug) 
				 {
				 console.log("Mapped InputDefinitions", jsonMessage);
				 };
				 

				 	postMessageToProxy($scope.proxyBuildViews, jsonMessage).then(function (message) {
					console.log("got message from proxy", message.response);
					$scope.loadStatus = $scope.layout.props.statusReadyLabel;
					$scope.lastEdit = new Date(); //set update to now
					$scope.$apply();
					});

			}
			
			function mapJsonToInputDefinitionConfiguration(json) {
			    //match the inpyt type to corresponding ENUM in the backend, used to ensure that the string can be matched to the enum values.
				const inputTypeMapping = {
					textbox: 'Textbox',
					text: 'Text',
					number: 'Number',
					checkbox: 'Checkbox',
					date: 'Date',
					select: 'Dropdown',
					radio: 'RadioButtonList',
					range: 'Slider'
				};
				
				if ($scope.layout.props.showDebug) { console.log("Original json object", json); }
				
				var convertedJson =  {
					Header: json.header,
					Dataid: json.id,
					Placeholder: json.placeholder || "",
					Type: inputTypeMapping[json.type.toLowerCase()] || json.type,
					InputValuesDefinition: (json.values || []).map(value => ({
						Value: value.value,
						Key: value.key
					})),
					InputOptionalAttributesDefinition: (json.optionalAttributes || []).map(attr => ({
						Attribute: attr.attribute,
						Value: attr.value
					})),
					InputAttributesDefinition: {
						Max: json.attributes.max || null,
						Min: json.attributes.min || null,
						Step: json.attributes.step || null,
						DefaultValue: json.defaultValue || null, 
						Placeholder: json.placeholder || "" 
					},
					Layout: {
						InputOrientationVertical: (json.layout && json.layout.inputOrientationVertical) || false
					},
					DynamicValueExpression: json.dynamicValueExpression || "",
					ValueType: json.valuesType || null,
					RegexValidationPattern: json.attributes["ng-pattern"] || ""
				};
				
				if ($scope.layout.props.showDebug) { console.log("Converted json object", convertedJson); }
				
				return convertedJson;
			}
			
			
			//update data - assumes valid input
			function updateData(rowKey, colmumnKey, input, rowData){
			
				$scope.loadStatus = $scope.layout.props.statusSavingLabel;
			
				//retrieve data from the row record
				var recordInfo = rowData.map(function (row) {

					//find title of the field
					if (row.qDimensionInfo && row.qDimensionInfo.qFallbackTitle) {
						var titleLabel = row.qDimensionInfo.qFallbackTitle;
					} else if (row.qMeasureInfo && row.qMeasureInfo.qFallbackTitle) {
						var titleLabel = row.qMeasureInfo.qFallbackTitle
					} else {
						var titleLabel = "unknown";
					}

					return { title: titleLabel, data: row.qText };

				});

				if ($scope.layout.props.showDebug) { console.log("recordInfo", recordInfo); }

				//set gui info for logging
				var guiInfo = { source: "QlikSense" }
				var keySeperator = $scope.layout.props.keySeperator;

				var jsonMessage = {
					key: rowKey + keySeperator + colmumnKey,
					input: input,
					recordInfo: JSON.stringify(recordInfo),
					guiInfo: JSON.stringify(guiInfo)
				};

				postMessageToProxy($scope.proxyUpdateComment, jsonMessage).then(function (message) {
					console.log("got message from proxy", message.response);
					$scope.loadStatus = $scope.layout.props.statusReadyLabel;
					$scope.lastEdit = new Date(); //set update to now
					$scope.$apply();
				});
			
			}

			//Helper functions

			function getDataFromProxy(url, keys) {

				if( $scope.layout.props.useUniqueKeys ) {
					// Use uniqueKeys
					return postGetDataFromUniqueKeys(url, keys)
				} else {
					// Use baseKey
					// keys param not used here
					return getdataFromBaseKey(url);
				}
			}
			
			function getdataFromBaseKey(url) {
				if ($scope.layout.props.showDebug) { console.log("getting data from", url); }
				return requestToProxyWithAuth(url, "GET").then(function (message) {
					if (!!message.response) {
						var data = JSON.parse(message.response);
						if (data.length > 0) {
							//console.log("Got data",data);
							return data;
						} else {
							console.log("Empty array found");
							return message.response;
						}
					} else {
						console.log("no data returned", message);
					}
				});
			}
			
			function postGetDataFromUniqueKeys(url, jsonBody) {
				if ($scope.layout.props.showDebug) { console.log("getting unique keys from", url); }
				
				//var jsonMessage = jsonBody;
				var jsonMessage = jsonBody;

				return postMessageToProxy(url, jsonMessage).then(function (message) {
					if (!!message.response) {
						var data = JSON.parse(message.response);
						if (data.length > 0) {
							if ($scope.layout.props.showDebug) { console.log("Got data",data); }
							return data;
						} else {
							console.log("Empty array found");
							return message.response;
						}
					} else {
						console.log("no data returned", message);
					}
				});

			}
			

			function postMessageToProxy(url, message) {
				return requestToProxyWithAuth(url, "POST", message);
			}


			function requestToProxyWithAuth(url, method, message) {
				//check if extension should handle auth
				if ($scope.layout.props.service.authentication.isEnabled && $scope.layout.props.service.authentication.isInteractive) {
					console.log("Request data with interactive auth");

					return requestToProxy(url, method, message).catch(function (error) {
						setMessagesToAuthenticating();
						console.log("Request error, trying to auth");
						return $scope.openAuthWindow().then(function (state) {
							clearErrorWidget(); //clear authentication message
							console.log("Request error, auth done retrying request");
							return requestToProxy(url, method, message);
						});;

					});
				} else {
					if ($scope.layout.props.showDebug) { console.log("Request without interactive auth"); }
					return requestToProxy(url, method, message);
				}
			}
			
			function makeRequestErrorFromStauts(xhr){
				
				if(!!xhr && !!xhr.status && xhr.status == 404){
					return makeError("Error communicating with Writebase service", "The service could not be contacted, please check settings. Status: " + xhr.status + " Statustext: " + xhr.statusText,"requestError");
				}
				
				if(!!xhr && !!xhr.status && xhr.status == 401){
					return makeError("Error communicating with Writebase service", "The service failed the authentication, please check settings. Status: " + xhr.status + " Statustext: " + xhr.statusText,"authenticationError");
				}
				
				if(!!xhr && !!xhr.status && xhr.status == 500){
					return makeError("Error communicating with Writebase service", "The service failed to find the solution Id, please check settings. Status: " + xhr.status + " Statustext: " + xhr.statusText,"authenticationError");
				}
				
				return makeError("Error communicating with Writebase service", "Status: " + xhr.status + " Statustext: " + xhr.statusText,"requestError");
			}

			//reference https://gomakethings.com/promise-based-xhr/
			function requestToProxy(url, method, message) {

				var xhr = new XMLHttpRequest();

				return new Promise(function (resolve, reject) {
					// Setup our listener to process compeleted requests
					xhr.onreadystatechange = function () {

						// Only run if the request is complete
						if (xhr.readyState !== 4) return;

						// Process the response
						if (xhr.status >= 200 && xhr.status < 300) {
							// If successful
							if ($scope.layout.props.showDebug) { console.log("Response", xhr); }
							resolve(xhr);
						} else {
							// If failed
							if ($scope.layout.props.showDebug) { console.log("request failed failed", xhr); }
							console.log("error",  xhr);
							
							//update scope load status
							$scope.loadStatus = "Request failed. For details go to developer tools or check log files.";
							
							reject(makeRequestErrorFromStauts(xhr));
						}
					};

					try {
						if (method === "GET") {
							xhr.open("GET", url, true);
							setRequestAuthentication(xhr);
							xhr.send();
						} else if (method == "POST") {
							xhr.open("POST", url, true);
							setRequestAuthentication(xhr);
							xhr.setRequestHeader('Content-Type', 'application/json');
							xhr.send(JSON.stringify(message));
						} else {
							console.error("unhandled method", method);
						}
					} catch (ex) {
						console.log("error", ex, xhr);
						reject(makeRequestErrorFromStauts(xhr));
					}

				});
			}

			//auth functions

			function setRequestAuthentication(xhr) {
				//confgiure request wiht the auth options
				if ($scope.layout.props.service.authentication.isEnabled) {
					if ($scope.layout.props.service.authentication.type == "NTLM") {
						xhr.withCredentials = true;
					} else if ($scope.layout.props.service.authentication.type == "HEADER") {
						xhr.withCredentials = true;
						xhr.setRequestHeader("QlikUser", "QlikUser " + getQlikUsername());
						xhr.setRequestHeader("apiKey",$scope.layout.props.service.authentication.apiKey);
					} else {
						console.error("unhandled auth type of", $scope.layout.props.service.authentication.type);
						reject("unhandled auth type of " + $scope.layout.props.service.authentication.type);
					}
				}
			}

			function getQlikUsername() {
				var currentUser = $scope.layout.user;
				if (currentUser.length > 1) {
					return currentUser;
				} else {
					throw ("user name is unexpected short: " + currentUser);
				}
			}
			
			function getRowKeyFromRow(row){			
				
				let keySegments = row.dimensions
					.filter(function(item){
					return !!item.qDimensionInfo && item.qDimensionInfo.partOfKey;
				})
				.map(function(item){
					return item.qText;
				});
				
				let keySeperator = $scope.layout.props.keySeperator;
				
				return keySegments.join(keySeperator);
			}
			$scope.getRowKeyFromRow = getRowKeyFromRow;
			
			function getKeyColumnIndex(dimensionsInfo) {
				let keyColumnsInformation = getKeyColumnsInformationFromDimensionsInfo(dimensionsInfo);
				let keyIndexs= keyColumnsInformation.map(function(item){
					return item.col;
				});
				return keyIndexs;
			}

			function getKeyColumnsInformationFromDimensionsInfo(dimensionsInfo){
				
				if(!!dimensionsInfo && dimensionsInfo.length > 0){
					let keyColumnsInformation = dimensionsInfo					
						.map(function(item, index){
							
								if(!!item && item.partOfKey !== undefined ){
									let header = {
										col: index,
										dimensionInfo: item								
									};

									return header;
								}
							})
						.filter(function(item){
							return !!item && !!item.dimensionInfo && item.dimensionInfo.partOfKey;
						});					
					
					return keyColumnsInformation;
					
				} else {
					console.log("##dimensionInfos is empty");
				}			
			}
			
			
			
			//extract the keys and add to list
			function getKeysFromDataPages(dataPages, keyColumns) {		
				
				//get contact from dataPages
				var keys = [];
				//var numberOfColumns = $scope.numberOfColumns;
				
				let keySeperator = $scope.layout.props.keySeperator;

				//check datapages
				if (dataPagesHasValidData(dataPages)) {

					//loop data pages
					//dataPages.forEach(function (dataPage) {
					for(var dataPage of dataPages) {
						//check data page
						if (!!dataPage && !!dataPage.qMatrix && dataPage.qMatrix.length > 0) {
							//loop rows in page
							//dataPage.qMatrix.forEach(function (row) {
							for(var row of dataPage.qMatrix){
								//check row
								if (!!row && row.length > 0) {
								
									//check key columns
									if(!!keyColumns && keyColumns.length > 0){
										let keyColumnValues = [];

									//get values from the key columens
								//	row.forEach(function(item, index){
								//		if(keyColumns.indexOf(index) != -1 ){
								//			keyColumnValues.push(row[index].qText);
								//		}
								//	});
										
									for(var keyColumn = 0; keyColumn <= row.length; keyColumn++){
										if(keyColumns.indexOf(keyColumn) != -1 ){
											keyColumnValues.push(row[keyColumn].qText);
										}
									}

									
										let key = keyColumnValues.join(keySeperator);
										keys.push(key);
									} else {
										console.log("unexpected keyColumns", keyColumns);
									}
									
								} else {
									console.log("unexpected row", row);
								}
							}//});
						} else {
							console.log("unexpected dataPage", dataPage);
						}
					}//});

				} else {
					console.log("Unexpected datapages", dataPages);
				}
				return keys;
			}

			function dataPagesHasValidData(dataPages) {
				//check datapages
				if (!!dataPages && dataPages.length > 0) {

					//check first datapage
					if (!!dataPages[0] && !!dataPages[0].qMatrix && dataPages[0].qMatrix.length > 0) {
						return true;
					} else {
						console.log("Empty dataset found in dataPages", dataPages);
					}
				} else {
					console.log("Datapages in empty or null", dataPages);
				}
				return false;
			}
			
			function getKeyNumberOfElements(keys){
				let keySeperator = $scope.layout.props.keySeperator;
				if(!!keys && keys.length > 0){
					var firstKey = keys[0].split(keySeperator);
					var lastKey = keys[keys.length - 1].split(keySeperator);
					if (firstKey.length == lastKey.length) {

						return firstKey.length;
					}	
				} else {
					throw new Error("No keys was found, please ensure that the some dimensions is set to partOfKey");
				}
			}

			//check keys for only mixed dimension, only the last element must change
			function checkKeysForMixedKeys(keys) {
				//var keyElements = $scope.layout.props.keyElements;
				if (keys.length >= 1) {
					let keySeperator = $scope.layout.props.keySeperator;
					var firstKey = keys[0].split(keySeperator);
					var lastKey = keys[keys.length - 1].split(keySeperator);

					if (firstKey.length == lastKey.length) {
					
					var keyElements = getKeyNumberOfElements(keys);

						for (var i = 0; i < keyElements - 1; i++) {
							if (firstKey[i] != lastKey[i]) {
								console.log("mixed keys in table, not supported", firstKey, lastKey);
								throw new Error("mixed keys in table, not supported first key is " + keys[0] + " last key is " + keys[keys.length - 1] +"")
								return false;
							}
						}
						//all good
						return true;

					} else {
						//console.log("key is missing expected seperator '|' and/or is too short");
						throw new Error("Key has an unexpected length, expected: " + keyElements + " got: " + lastKey.length);
					}

				} else {
					//throw new Error("No keys in table, check the data");
					console.log("no keys in table");
				}
				return false;
			}

			function getBaseKey(keys) {
				var keyElements = getKeyNumberOfElements(keys);
				var keySepetrator = $scope.layout.props.keySeperator;

				if (checkKeysForMixedKeys(keys)) {
					var firstKey = keys[0].split(keySepetrator);
					var baseKey = firstKey.slice(0, keyElements - 1).join(keySepetrator);
					return baseKey;
				} else {
					console.log("Keys not valid");
					throw new Error("Keys not valid");
				}
			}

			function findLastestTimestampIndata(data) {
				var actionKey = $scope.layout.props.actionKey;
				return Math.max.apply(Math,
					data
						.filter(function (inputRecord) { return !inputRecord.key.endsWith(actionKey) }) //remove actionKey entries from list
						.map(function (item) { return new Date(formatDateToUTC(item.createtime)); })
				);
			}
			
			function formatDateToUTC(dateString) {
			  if (dateString.charAt(dateString.length - 1) !== 'Z') {
				return dateString + 'Z';
			  }
			  return dateString;
			}

			function dbDateToDate(date) {
				var dateString = date.toString();

				var year = dateString.substring(0, 4);
				var month = dateString.substring(4, 6);
				var day = dateString.substring(6, 8);
				var hour = dateString.substring(8, 10);
				var minute = dateString.substring(10, 12);
				var second = dateString.substring(12, 14);

				//console.log(year,month,day,hour,minute,second);

				return new Date(year, month - 1, day, hour, minute, second);
			}

			function getLastActionTimestamp(baseKey, data) {
				//get definitions
				var inputDefinitions = angular.copy($scope.inputDefinition);
				var keySepetrator = $scope.layout.props.keySeperator;
				var actionKey = $scope.layout.props.actionKey;

				//look for the records with the actionKey
				var inputData = data.filter(function (inputRecord) { return inputRecord.key == baseKey + keySepetrator + actionKey });

				if (!!inputData && inputData.length == 1) {
					var inputRecord = inputData[0];
					if ($scope.layout.props.showDebug) { console.log("found input data", inputRecord); }
					return dbDateToDate(inputRecord.createtime);
				} else {
					if ($scope.layout.props.showDebug) { console.log("found no or multiple records, expects one "); }
					return null;
				}
			}
			
			//// POLYFILLS /////

			//fixing IE11
			if (!String.prototype.endsWith) {
				String.prototype.endsWith = function (searchString, position) {
					var subjectString = this.toString();
					if (typeof position !== 'number' || !isFinite(position)
						|| Math.floor(position) !== position || position > subjectString.length) {
						position = subjectString.length;
					}
					position -= searchString.length;
					var lastIndex = subjectString.indexOf(searchString, position);
					return lastIndex !== -1 && lastIndex === position;
				};
			}

			// Production steps of ECMA-262, Edition 6, 22.1.2.1
			//https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/from
			if (!Array.from) {
				Array.from = (function () {
					var symbolIterator;
					try {
						symbolIterator = Symbol.iterator
							? Symbol.iterator
							: 'Symbol(Symbol.iterator)';
					} catch (e) {
						symbolIterator = 'Symbol(Symbol.iterator)';
					}

					var toStr = Object.prototype.toString;
					var isCallable = function (fn) {
						return (
							typeof fn === 'function' ||
							toStr.call(fn) === '[object Function]'
						);
					};
					var toInteger = function (value) {
						var number = Number(value);
						if (isNaN(number)) return 0;
						if (number === 0 || !isFinite(number)) return number;
						return (number > 0 ? 1 : -1) * Math.floor(Math.abs(number));
					};
					var maxSafeInteger = Math.pow(2, 53) - 1;
					var toLength = function (value) {
						var len = toInteger(value);
						return Math.min(Math.max(len, 0), maxSafeInteger);
					};

					var setGetItemHandler = function setGetItemHandler(isIterator, items) {
						var iterator = isIterator && items[symbolIterator]();
						return function getItem(k) {
							return isIterator ? iterator.next() : items[k];
						};
					};

					var getArray = function getArray(
						T,
						A,
						len,
						getItem,
						isIterator,
						mapFn
					) {
						// 16. Let k be 0.
						var k = 0;

						// 17. Repeat, while k < len… or while iterator is done (also steps a - h)
						while (k < len || isIterator) {
							var item = getItem(k);
							var kValue = isIterator ? item.value : item;

							if (isIterator && item.done) {
								return A;
							} else {
								if (mapFn) {
									A[k] =
										typeof T === 'undefined'
											? mapFn(kValue, k)
											: mapFn.call(T, kValue, k);
								} else {
									A[k] = kValue;
								}
							}
							k += 1;
						}

						if (isIterator) {
							throw new TypeError(
								'Array.from: provided arrayLike or iterator has length more then 2 ** 52 - 1'
							);
						} else {
							A.length = len;
						}

						return A;
					};

					// The length property of the from method is 1.
					return function from(arrayLikeOrIterator /*, mapFn, thisArg */) {
						// 1. Let C be the this value.
						var C = this;

						// 2. Let items be ToObject(arrayLikeOrIterator).
						var items = Object(arrayLikeOrIterator);
						var isIterator = isCallable(items[symbolIterator]);

						// 3. ReturnIfAbrupt(items).
						if (arrayLikeOrIterator == null && !isIterator) {
							throw new TypeError(
								'Array.from requires an array-like object or iterator - not null or undefined'
							);
						}

						// 4. If mapfn is undefined, then let mapping be false.
						var mapFn = arguments.length > 1 ? arguments[1] : void undefined;
						var T;
						if (typeof mapFn !== 'undefined') {
							// 5. else
							// 5. a If IsCallable(mapfn) is false, throw a TypeError exception.
							if (!isCallable(mapFn)) {
								throw new TypeError(
									'Array.from: when provided, the second argument must be a function'
								);
							}

							// 5. b. If thisArg was supplied, let T be thisArg; else let T be undefined.
							if (arguments.length > 2) {
								T = arguments[2];
							}
						}

						// 10. Let lenValue be Get(items, "length").
						// 11. Let len be ToLength(lenValue).
						var len = toLength(items.length);

						// 13. If IsConstructor(C) is true, then
						// 13. a. Let A be the result of calling the [[Construct]] internal method
						// of C with an argument list containing the single item len.
						// 14. a. Else, Let A be ArrayCreate(len).
						var A = isCallable(C) ? Object(new C(len)) : new Array(len);

						return getArray(
							T,
							A,
							len,
							setGetItemHandler(isIterator, items),
							isIterator,
							mapFn
						);
					};
				})();
			}
			
			/* - Element.prototype.replaceWith, License: CC0 */
			(function(self, undefined) {

			// _mutation
			var _mutation = (function () { // eslint-disable-line no-unused-vars

				function isNode(object) {
					// DOM, Level2
					if (typeof Node === 'function') {
						return object instanceof Node;
					}
					// Older browsers, check if it looks like a Node instance)
					return object &&
						typeof object === "object" && 
						object.nodeName && 
						object.nodeType >= 1 &&
						object.nodeType <= 12;
				}

				// http://dom.spec.whatwg.org/#mutation-method-macro
				return function mutation(nodes) {
					if (nodes.length === 1) {
						return isNode(nodes[0]) ? nodes[0] : document.createTextNode(nodes[0] + '');
					}

					var fragment = document.createDocumentFragment();
					for (var i = 0; i < nodes.length; i++) {
						fragment.appendChild(isNode(nodes[i]) ? nodes[i] : document.createTextNode(nodes[i] + ''));

					}
					return fragment;
				};
			}());

			// Element.prototype.replaceWith
			/* global _mutation */
			Document.prototype.replaceWith = Element.prototype.replaceWith = function replaceWith() {
				if (this.parentNode) {
					this.parentNode.replaceChild(_mutation(arguments), this);
				}
			};

			// Not all UAs support the Text constructor.  Polyfill on the Text constructor only where it exists
			// TODO: Add a polyfill for the Text constructor, and make it a dependency of this polyfill.
			if ('Text' in self) {
				Text.prototype.replaceWith = Element.prototype.replaceWith;
			}
			})
			('object' === typeof window && window || 'object' === typeof self && self || 'object' === typeof global && global || {});


		}]
	};
	
	// Add directives to qvAngular
	function addAppDirectives(app){

		/*app.directive('includeReplace', function () {
			return {
				require: 'ngInclude',
				restrict: 'A',
				link: function (scope,tElement, tAttrs) {
				console.log("element",tElement,"attr",tAttrs, "children", tElement.children());
					tElement.replaceWith(tElement.children());
					return {
						post : angular.noop
					};
				}
			};
		});*/

		app.directive("directiveWhenScrolled", function () {
			return function (scope, elm, attr) {
				var raw = elm[0];

				elm.bind('scroll', function () {
					if (Math.ceil(raw.scrollTop) + raw.offsetHeight >= raw.scrollHeight) {
						scope.$apply(attr.directiveWhenScrolled);
					}
				});
			};
		});
	}
});
