/*global define*/
define( [
	'jquery',
	'qvangular',
	'',
	''

], function ( $, qvangular, ngTemplate, cssContent ) {
	'use strict';

	$( "<style>" ).html( cssContent ).appendTo( "head" );

	qvangular.directive('stretchDynamicInput', function($compile) {
    
	   
		function arrayEquals(a, b) {
		  return Array.isArray(a) &&
			Array.isArray(b) &&
			a.length === b.length 
			&& a.every(function(val, index) { 
				if(Array.isArray(val) || Array.isArray(b[index])){				
					return arrayEquals(val, b[index]);
				} else {
					return JSON.stringify(val) === JSON.stringify(b[index]); 
				}
			})
			;
		}	
		
		function objectEquels(a,b){
			return JSON.stringify(a) === JSON.stringify(b);
		}
		
		 

	   function typeToTag(definition, defaultPlaceholder, inputTabOrder){
			if(definition.type == "checkbox"){
				return  [
					{
						attributes: 'type="checkbox" ng-change="onInputChangeFn(inputDataRow,inputDataColumnKey,model, inputDataRow.cells)"  ng-true-value="\'true\'" ng-true-value="\'false\'"',
						defaultAttributes: {},						
						tag: 'input',
						type: 'input',
					}
					]; 
			} if(definition.type == "date"){
				return  [
					{
						attributes: 'type="date" ng-change="onInputChangeFn(inputDataRow,inputDataColumnKey, model, inputDataRow.cells)" ng-model-options="{timezone: \'UTC\'}"',
						defaultAttributes: {						
						},					
						tag: 'input',
						type: 'input',
					}
					];
			} if(definition.type == "range"){
				return  [
					{
						attributes: 'type="range" ng-blur="onBlur()"  ng-focus="onFocus()" title="{{model}}" ',
						defaultAttributes: {						
						},					
						tag: 'input',
						type: 'input',
					},
					{
						attributes: '',
						tag: 'label',
						value: '{{model}}',
					}
					];
			} if(definition.type == "radio"){
				
				var elements = [];
				
				definition.values.forEach(function(value, index){
					//add control
					elements.push({
						attributes: 'type="radio" ng-change="onInputChangeFn(inputDataRow,inputDataColumnKey,model, inputDataRow.cells)" title="{{model}}" value="' + value.value +'" id="'+inputTabOrder+index+'"',
						defaultAttributes: {						
						},					
						tag: 'input',
						type: 'input',
						
					});
					
					//add label
					elements.push({
						attributes: ' for="'+inputTabOrder+index+'"',
						tag: 'label',
						type: 'layout',
						value: value.value,
					});
					
					if(!!definition.layout && !!definition.layout.inputOrientationVertical && definition.layout.inputOrientationVertical == true){
						//add label
						elements.push({
							attributes: '',
							tag: 'br',
							type: 'layout',
						});
					}
				
				
				});
				return elements;
				
			} if(definition.type == "select"){
				return  [
					{
						attributes: 'ng-change="onInputChangeFn(inputDataRow,inputDataColumnKey,model, inputDataRow.cells)"',
						defaultAttributes: {						
						},					
						tag: 'select',
						type: 'input',
						children: {
							tag: 'option',
							values: definition.values
							}
						}
					];
			} if(definition.type == "textbox"){
				return [
					{
						attributes: 'ng-keyup="autoExpandFn($event.target)" ng-blur="onBlur()" ng-focus="onFocus()"', 
						defaultAttributes: {
							rows:'1', 
							cols:'20',
							placeholder: defaultPlaceholder
						},
						tag: 'textarea',
						type: 'input',
					}
					];	
			} if(definition.type == "number"){
				return  [
					{
						attributes: 'type="number" ng-blur="onBlur()"  ng-focus="onFocus()" ng-model-options="{ allowInvalid: true}"',
						defaultAttributes: {						
						},					
						tag: 'input',
						type: 'input'	
					}
					];
			} else {
				//text af default
				return  [
					{
						attributes: 'type="text" ng-blur="onBlur()"  ng-focus="onFocus()"',
						defaultAttributes: {						
						},					
						tag: 'input',
						type: 'input'	
					}
					];	 //default
								
			}	   
	   };
	   
	   function template(model, definition, defaultPlaceholder, inputTabOrder) {
	   
	   
	   
	   		//console.log("definition",definition);

			//start element
			var elementTypes = typeToTag(definition, defaultPlaceholder, inputTabOrder);
			
			var templateString = "";
			
			//loop defined elements
			elementTypes.forEach(function(elementType){
				
				//set elemettype
				templateString += "<" + elementType.tag;
				

				//set default properties
				templateString += " " + elementType.attributes;

				//add default attributes if not set in config
				for (var defaultAttribute in elementType.defaultAttributes){
					if(!definition.attributes[defaultAttribute]){
						if (defaultAttribute != "$$hashKey" ) {
							templateString += " " + defaultAttribute + '="' + elementType.defaultAttributes[defaultAttribute] + '"';
						}
					}				
				}
				
				//add datamodel link and tab order to inputs
				if(!!elementType.type && elementType.type == "input"){

					//add options attributes
					if(!!definition.optionalAttributes && Array.isArray(definition.optionalAttributes)){
						definition.optionalAttributes.forEach( function(optionalAttribute){	
							if(!!optionalAttribute && !!optionalAttribute.attribute && !!optionalAttribute.value){
								templateString += " " + optionalAttribute.attribute + '="' + optionalAttribute.value + '"';
							}
						});
					}

					//add dynamic attributes
					for (var attribute in definition.attributes) {
						if (attribute != "id" && attribute != "header" && attribute != "type" && attribute != "values" && attribute != "$$hashKey" && attribute != "attributes" ) {
							templateString += " " + attribute + '="' + definition.attributes[attribute] + '"';
						}
					}
				

					//add name
					templateString += ' name="' + inputTabOrder + '"';

					//add ng-model

					templateString += ' ng-model="model"';


					//add tab order
					templateString += ' ng-attr-tabIndex="'+inputTabOrder+'"';
					
					//add element disabled
					templateString += ' ng-disabled="inputDisabled"';
				}
				
				//handle single type elements
				if(elementType.tag == 'br'){
					templateString += "/>";
					return;
				}


				templateString += ">";

				//add chilren
				if(!!elementType.children &&  !!elementType.children.values){
					for (var childValue in elementType.children.values){
						templateString += '<'+elementType.children.tag;
						templateString += ' value="'+ elementType.children.values[childValue].key +'"';
						templateString += ' >' + elementType.children.values[childValue].value					
						templateString += '</' + elementType.children.tag + '>';				
						}
				}
				
				if(!!elementType.value){
					templateString += elementType.value;
				}

				templateString += "</" + elementType.tag + ">";
			
			
			});
			

			//console.log(templateString);
			return templateString;
		};
		
		function createElement(scope){
			var html = template(scope.model,scope.definition,scope.defaultPlaceholder, scope.inputTabOrder);
			var dynamicElement = $compile(html)(scope);
			return dynamicElement;
		}

		return {
			restrict: "E",
			scope: {
				model :'=ngModel',
			  	definition : '=',
				defaultPlaceholder : '<', //TODO add check for custom placeholder attribute
				inputDataRow : '=',
				inputDataColumnKey : '=',
				inputDisabled : '=',
				inputTabOrder : '@',
			  	autoExpand : '&',
				onBlur : '&',
				onFocus : '&',
				onInputChange : '&',
			},
			transclude : true,
			link: function(scope, element, attrs){
				
				//unwrapping functions
				scope.autoExpandFn = function(obj){
					scope.autoExpand()(obj);
				}
								
				scope.onInputChangeFn = function(row, colmumnKey, input, rowData, oldValue){
					scope.onInputChange()(row, colmumnKey, input, rowData, oldValue);
				}

				
				scope.$watchGroup(['definition.type','definition.values','definition.placeholder','definition.attributes','definition.layout'], function (newval, oldval) {
					//console.log("new definition in component values", newval, oldval, arrayEquals(newval, oldval), element);
					if(!objectEquels(newval, oldval)){
						
						//clear
						element.empty();
						//clear
						element.append(createElement(scope));
					}

				});

				//add initial element
				element.append(createElement(scope));
			}

		};

	});

} );